/*
 * arch.c
 *
 *  Created on: 27.04.2009
 *      Author: Alexander Boldyrev
 */
#include <io.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/utime.h>
#include <time.h>
#include <fcntl.h>

#include "local.h"

typedef struct FileHeader {
	char *name;
	unsigned long size;
	unsigned long ctime;
	unsigned long mtime;
	unsigned long hash;
} FileHeader;

#define ALPHABET_SIZE_LOG_2 8
#define ALPHABET_SIZE 256
#define BLOCK_SIZE 4096
#define WARNING_FILE_NAME "__warning__"

typedef struct HaffmanTree {
	struct HaffmanTree *left, *right;
	unsigned long character;
	unsigned long weight;
} HaffmanTree;

static unsigned long getLowestTree(HaffmanTree *haffmanTree, unsigned long used) {
	unsigned long i, min = 0;

	for (i = 1; i < used; i++) {
		if (haffmanTree[i].weight < haffmanTree[min].weight) {
			min = i;
		}
	}

	return min;
}

static void traverse(HaffmanTree *tree, unsigned char *code,
		unsigned char count, unsigned char *codes) {
	if (tree->character) {
		int i;
		for (i = 0; i < ALPHABET_SIZE_LOG_2; i++) {
			codes[tree->character * (ALPHABET_SIZE_LOG_2 + 1) + i] = code[i];
		}
		codes[tree->character * (ALPHABET_SIZE_LOG_2 + 1) + ALPHABET_SIZE_LOG_2]
				= count;
	}
	if (tree->left) {
		code[count] = 0;
		traverse(tree->left, code, count + 1, codes);
	}
	if (tree->right) {
		code[count] = 1;
		traverse(tree->right, code, count + 1, codes);
	}
}

static unsigned char *compress(unsigned char *buffer, unsigned long *size) {
	unsigned char *result = NULL;
	unsigned long i, used, tempUsed, newSize, j;
	unsigned long weights[ALPHABET_SIZE];
	unsigned char codes[(ALPHABET_SIZE_LOG_2 + 1) * ALPHABET_SIZE];
	unsigned char code[ALPHABET_SIZE_LOG_2], byte;
	HaffmanTree haffmanTree[ALPHABET_SIZE];

	memset(weights, 0, ALPHABET_SIZE * sizeof(unsigned long));
	for (i = 0; i < *size; i++) { // Compute weights for all bytes
		weights[buffer[i]]++;
	}

	//
	// Build leaf nodes
	//
	used = 0;
	for (i = 0; i < ALPHABET_SIZE; i++) {
		int w = weights[i];
		if (w != 0) {
			haffmanTree[used].character = i;
			haffmanTree[used].weight = w;
			haffmanTree[used].left = haffmanTree[used].right = NULL;
			used++;
		}
	}
	tempUsed = used;

	//
	// Build tree
	//
	while (used > 1) {
		int min = getLowestTree(&haffmanTree[0], used);
		int weight = haffmanTree[min].weight;
		HaffmanTree temp;
		temp.left = malloc(sizeof(HaffmanTree));
		*temp.left = haffmanTree[min];
		haffmanTree[min] = haffmanTree[--used];

		min = getLowestTree(&haffmanTree[0], used);
		temp.right = malloc(sizeof(HaffmanTree));
		*temp.right = haffmanTree[min];
		temp.weight = weight + haffmanTree[min].weight;
		haffmanTree[min] = temp;
	}

	memset(code, 0, ALPHABET_SIZE_LOG_2 * sizeof(unsigned char));
	memset(codes, 0, (ALPHABET_SIZE_LOG_2 + 1) * ALPHABET_SIZE
			* sizeof(unsigned char));
	traverse(&haffmanTree[0], code, 0, codes);

#if 0
	printf("\n");
	for (i = 0; i < ALPHABET_SIZE; i++) {
		unsigned char j;
		if (weights[i] == 0) {
			continue;
		}
		printf("%d :  ", i);
		for (j = 0; j < codes[i * (ALPHABET_SIZE_LOG_2 + 1) + 8]; j++) {
			printf("%d ", codes[i * (ALPHABET_SIZE_LOG_2 + 1) + j]);
		}
		printf("\n");
	}
	printf("\n");
#endif

	used = tempUsed;
	newSize = 5 + 3 * used;

	result = malloc(newSize);
	result[0] = tempUsed % 256;
	result[1] = (*size) % 256;
	result[2] = (*size / 256) % 256;
	result[3] = (*size / 256 / 256) % 256;
	result[4] = (*size / 256 / 256 / 256) % 256;
	//
	// Save information about compressing for decoder
	//
	for (i = 0, j = 0; i < ALPHABET_SIZE; i++) {
		char *code = &codes[i * (ALPHABET_SIZE_LOG_2 + 1)];
		unsigned short value = 128, f;
		if (weights[i] == 0) {
			continue;
		}
		result[5 + j * 3 + 0] = i;

		if (code[8] == 0 && used == 1) { // only one character
			code[8] = 1;
		}

		result[5 + j * 3 + 1] = 0;
		for (f = 0; f < code[8]; f++) {
			result[5 + j * 3 + 1] += code[f] * value;
			value /= 2;
		}
		result[5 + j * 3 + 2] = code[8];

		j++;
	}
	//
	// Compress data
	//
	for (i = 0, j = 0, byte = 0; i < *size; i++) {
		char *code = &codes[buffer[i] * (ALPHABET_SIZE_LOG_2 + 1)];
		unsigned char tempByte;
		unsigned short value = 128, f;

		tempByte = 0;
		for (f = 0; f < code[8]; f++) {
			tempByte += code[f] * value;
			value /= 2;
		}

		if (j + code[8] > 8) {
			byte = byte + (tempByte >> j);

			result = realloc(result, ++newSize);
			result[newSize - 1] = byte;

			byte = tempByte << (8 - j);
			j = code[8] - (8 - j);
		} else {
			byte = byte + (tempByte >> j);
			j += code[8];
		}
	}

	result = realloc(result, ++newSize);
	result[newSize - 1] = byte;

	*size = newSize;
	free(buffer);
	return result;
}
static unsigned char *decompress(unsigned char *buffer, unsigned long *size) {
	unsigned char used = buffer[0], byte;
	unsigned char codes[(ALPHABET_SIZE_LOG_2 + 2) * ALPHABET_SIZE];
	unsigned long newSize;
	unsigned long i, f, offset, offsetResult = 0;
	unsigned char *result;

	newSize = buffer[1] + 256 * buffer[2] + 256 * 256 * buffer[3] + 256 * 256
			* 256 * buffer[4];
	result = malloc(newSize);
	if (!result) {
		return NULL;
	}

	memset(codes, 0, sizeof(codes));
	for (i = 0; i < used; i++) {
		char *code = &codes[i * (ALPHABET_SIZE_LOG_2 + 2)];

		for (f = 0; f < ALPHABET_SIZE_LOG_2; f++) {
			code[f] = (buffer[5 + i * 3 + 1] << f) >> (ALPHABET_SIZE_LOG_2 - 1);
		}
		code[8] = buffer[5 + i * 3 + 2];
		code[9] = buffer[5 + i * 3];
	}
	offset = 5 + used * 3;

	byte = buffer[offset];
	i = 0;
	while (1) {
		unsigned char code[10];
		unsigned char done = 0;
		code[8] = 0;
		while (!done && i < 8) {
			code[code[8]] = (byte << code[8]) >> (ALPHABET_SIZE_LOG_2 - 1);
			code[8]++;

			for (f = 0; f < used && !done; f++) {
				unsigned char *tempCode = &codes[f * (ALPHABET_SIZE_LOG_2 + 2)];
				unsigned char j;

				if (tempCode[8] != code[8]) {
					continue;
				}
				done = 1;
				for (j = 0; j < code[8]; j++) {
					if (code[j] != tempCode[j]) {
						done = 0;
						break;
					}
				}

				if (done) {
					code[9] = tempCode[9];
				}
			}
		}
		result[offsetResult++] = code[9];

		if (offsetResult == newSize) {
			break;
		}

		i += code[8];
		if (i > 7) {
			i -= 8;

			byte = buffer[++offset] << i;
			if (offset < (*size) - 1) {
				byte += buffer[offset + 1] >> (ALPHABET_SIZE_LOG_2 - i);
			}

		} else {
			byte = byte << code[8];
			if (offset < (*size) - 1) {
				byte += buffer[offset + 1] >> (ALPHABET_SIZE_LOG_2 - code[8]);
			}
		}
	}

	free(buffer);
	*size = newSize;
	return result;
}

const char *archMakeTemp(const char *archFileName) {
	FILE *file = fopen(archFileName, "rb");
	FILE *temp = NULL;
	char *tempFileName = NULL;
	int i = 0;

	do { // Make temp file name
		int size = 0, tempI = i;

		if (temp) {
			fclose(temp);
		}

		if (i == 256) {
			return NULL; // For not cycling
		}

		if (tempI == 0) {
			size = 1;
		} else {
			while (tempI) {
				tempI /= 10;
				size++;
			}
		}
		if (tempFileName) {
			free(tempFileName);
		}
		tempFileName = malloc(strlen(archFileName) + strlen(".swp") + size + 1);
		sprintf(tempFileName, "%s.swp%d", archFileName, i);

		temp = fopen(tempFileName, "rb");

		i++;
	} while (temp);

	temp = fopen(tempFileName, "wb");

	if (file) { // Copy old archive to temp
		int size;
		unsigned char buffer[BLOCK_SIZE];

		while ((size = fread(buffer, 1, BLOCK_SIZE, file)) > 0) {
			if (fwrite(buffer, size, 1, temp) != 1) {
				return NULL;
			}
		}
		fclose(file);
	} else { // Create null archive as temp
		unsigned char byte = 0x00;
		fwrite(&byte, 1, 1, temp);
	}
	fclose(temp);

	return tempFileName;
}

static int writeULong(unsigned long value, FILE *file) {
	unsigned char byte[4];

	byte[0] = (value) % 256;
	byte[1] = (value / 256) % 256;
	byte[2] = (value / 256 / 256) % 256;
	byte[3] = (value / 256 / 256 / 256) % 256;
	if (fwrite(&byte[0], 1, 1, file) != 1) {
		return 1;
	}
	if (fwrite(&byte[1], 1, 1, file) != 1) {
		return 1;
	}
	if (fwrite(&byte[2], 1, 1, file) != 1) {
		return 1;
	}
	if (fwrite(&byte[3], 1, 1, file) != 1) {
		return 1;
	}

	return 0;
}
static int readULong(unsigned long *value, FILE *file) {
	unsigned char byte[4];

	if (fread(&byte[0], 1, 1, file) != 1) {
		return 1;
	}
	if (fread(&byte[1], 1, 1, file) != 1) {
		return 1;
	}
	if (fread(&byte[2], 1, 1, file) != 1) {
		return 1;
	}
	if (fread(&byte[3], 1, 1, file) != 1) {
		return 1;
	}

	*value = byte[0] + byte[1] * 256 + byte[2] * 256 * 256 + byte[3] * 256
			* 256 * 256;

	return 0;
}

int archAddFile(const char *archFileName, const char *fileName) {
	int input = open(fileName, O_RDONLY | O_BINARY);
	FILE *arch = fopen(archFileName, "rb+");

	unsigned char *inputData;

	struct stat stat_buf;
	FileHeader header;

	unsigned char byte;

	if (strcmp(fileName, WARNING_FILE_NAME) == 0) {
		return 0;
	}

	if (input < 0 || !arch) {
		return 1;
	}

	//
	// Load file and information about it to memory
	//
	if (fstat(input, &stat_buf) < 0) {
		return 1;
	}

	header.name = strdup(fileName);
	header.size = stat_buf.st_size;
	header.ctime = stat_buf.st_ctime;
	header.mtime = stat_buf.st_mtime;

	if (header.size <= 0) { // file is null
		printf("File \"%s\" is null. Skip it.\n", fileName);
		return 0;
	}

	inputData = malloc(header.size);

	if (!inputData || !header.name) {
		return 1;
	}
	if (read(input, inputData, header.size) != header.size) {
		return 1;
	}

	close(input);

	// Compress 'inputData' (Haffman's algorithm)
	inputData = compress(inputData, &header.size);
	header.hash = crc32(inputData, header.size);

	//
	// Write all to end of archive
	//

	// Seek to end of file without last zero
	if (fseek(arch, -1, SEEK_END)) {
		return 1;
	}

	byte = 0x01; // Write last zero
	if (fwrite(&byte, 1, 1, arch) != 1) {
		return 1;
	}
	if (fwrite(header.name, 1, strlen(header.name) + 1, arch) != (strlen(
			header.name) + 1)) {
		return 1;
	}

	// Write in this order: size, ctime, mtime and hash.
	if (writeULong(header.size, arch) || writeULong(header.ctime, arch)
			|| writeULong(header.mtime, arch) || writeULong(header.hash, arch)) {
		return 1;
	}

	if (fwrite(inputData, header.size, 1, arch) != 1) {
		return 1;
	}

	byte = 0x00; // Write last zero
	if (fwrite(&byte, 1, 1, arch) != 1) {
		return 1;
	}

	if (fclose(arch) == EOF) {
		return 1;
	}

	free(header.name);
	free(inputData);
	return 0;
}

int archMake(const char *archFileName, const char *tempFileName) {
	FILE *input = fopen(tempFileName, "rb");
	FILE *output = fopen(archFileName, "wb");
	unsigned long size;
	unsigned char buffer[BLOCK_SIZE];

	if (!input || !output) {
		return 1;
	}

	//
	// Copy all bytes from temp archive to main
	//

	while ((size = fread(buffer, 1, BLOCK_SIZE, input)) > 0) {
		if (fwrite(buffer, size, 1, output) != 1) {
			return 1;
		}
	}

	if (fclose(input) == EOF) {
		return 1;
	}
	if (fclose(output) == EOF) {
		return 1;
	}

	remove(tempFileName);
	free((void *) tempFileName);

	return 0;
}

int archCheckFileNames(const char *archFileName, size_t count, char **names) {
	size_t countOfWarnings = 0;
	int i, j;

	// Delete names which is bad (see below)
	for (i = 1; i < count; i++) {
		if (strcmp(names[i], WARNING_FILE_NAME) == 0) { // Current name is already unuseable
			continue;
		}

		if (strcmp(names[0], names[i]) == 0) { // Current name coincides with archive name
			printf(
					"Warning: \"%s\" - bad name, couldn't use for compress itself.\n",
					names[0]);
			free(names[i]);
			names[i] = strdup(WARNING_FILE_NAME);
			countOfWarnings++;
			continue;
		}
		for (j = i + 1; j < count; j++) {
			if (strcmp(names[i], names[j]) == 0) { // Two identical name
				printf("Warning: \"%s\" - two identical name.\n", names[i]);
				free(names[j]);
				names[j] = strdup(WARNING_FILE_NAME);
				countOfWarnings++;
			}
		}
	}

	{
		FILE *arch = fopen(archFileName, "rb");
		unsigned char byte;

		if (!arch) {
			return 1;
		}

		if (fread(&byte, 1, 1, arch) != 1) {
			return 1;
		}
		while (byte) {
			char nameByte;
			unsigned long size;
			char name[1024];
			size_t nameLength = 0;

			if (fread(&nameByte, 1, 1, arch) != 1) {
				return 1;
			}
			while (nameByte) {
				name[nameLength++] = nameByte;
				if (fread(&nameByte, 1, 1, arch) != 1) { // Seek name
					return 1;
				}
			}
			name[nameLength] = 0;
			for (i = 1; i < count; i++) {
				if (strcmp(names[i], WARNING_FILE_NAME) == 0) { // Current name is already unuseable
					continue;
				}
				if (strcmp(name, names[i]) == 0) { // Archive already contain file with this name
					printf(
							"Warning: \"%s\" - bad name, archive already contain file with this name\n",
							names[i]);
					free(names[i]);
					names[i] = strdup(WARNING_FILE_NAME);
					countOfWarnings++;
				}
			}

			readULong(&size, arch);
			if (fseek(arch, size + sizeof(unsigned long) * 3, SEEK_CUR) < 0) {
				return 1;
			}

			if (fread(&byte, 1, 1, arch) != 1) {
				return 1;
			}
		}

		if (fclose(arch) == EOF) {
			return 1;
		}
	}

	return 0;
}

unsigned long *archCreateOffsetTable(const char *archFileName, size_t count,
		char **names) {
	size_t countOfWarnings = 0;
	int i, j;
	unsigned long *offsets = malloc(count * sizeof(unsigned long));
	int offset;

	memset(offsets, 0, count * sizeof(unsigned long));
	// Delete names which is bad (see below)
	for (i = 1; i < count; i++) {
		if (strcmp(names[i], WARNING_FILE_NAME) == 0) { // Current name is already unuseable
			continue;
		}

		if (strcmp(names[0], names[i]) == 0) { // Current name coincides with archive name
			printf(
					"Warning: \"%s\" - bad name, couldn't use for compress itself.\n",
					names[0]);
			free(names[i]);
			names[i] = strdup(WARNING_FILE_NAME);
			countOfWarnings++;
			offsets[i] = 0;
			continue;
		}
		for (j = i + 1; j < count; j++) {
			if (strcmp(names[i], names[j]) == 0) { // Two identical name
				printf("Warning: \"%s\" - two identical name.\n", names[i]);
				free(names[j]);
				names[j] = strdup(WARNING_FILE_NAME);
				offsets[j] = 0;
				countOfWarnings++;
			}
		}
	}

	{
		FILE *arch = fopen(archFileName, "rb");
		unsigned char byte;

		if (!arch) {
			return NULL;
		}

		if (fread(&byte, 1, 1, arch) != 1) {
			return NULL;
		}
		offset = 1;
		while (byte) {
			char nameByte;
			unsigned long size;
			char name[1024];
			size_t nameLength = 0;

			offset++;
			if (fread(&nameByte, 1, 1, arch) != 1) {
				return NULL;
			}
			while (nameByte) {
				name[nameLength++] = nameByte;
				offset++;
				if (fread(&nameByte, 1, 1, arch) != 1) { // Seek name
					return NULL;
				}
			}
			name[nameLength] = 0;
			for (i = 1; i < count; i++) {
				if (strcmp(names[i], WARNING_FILE_NAME) == 0) { // Current name is already unuseable
					continue;
				}
				if (strcmp(name, names[i]) == 0) { // Archive already contain file with this name
					offsets[i] = offset;
				}
			}

			readULong(&size, arch);
			offset += size + sizeof(unsigned long) * 4;
			if (fseek(arch, size + sizeof(unsigned long) * 3, SEEK_CUR) < 0) {
				return NULL;
			}

			offset++;
			if (fread(&byte, 1, 1, arch) != 1) {
				return NULL;
			}
		}

		if (fclose(arch) == EOF) {
			return NULL;
		}
	}
	for (i = 1; i < count; i++) {
		if (offsets[i] == 0) {
			printf("File %s not found in archive", names[i]);
		}
	}

	return offsets;
}

int archExtractFile(const char *archFileName, const char *fileName,
		unsigned long offset) {
	FILE *arch = fopen(archFileName, "rb");
	FILE *file = NULL;
	FileHeader header;

	unsigned char *buffer;

	if (offset == 0 || strcmp(fileName, WARNING_FILE_NAME) == 0) {
		return 0;
	}

	file = fopen(fileName, "wb");

	if (!arch || !file) {
		return 1;
	}

	// Seek to the file
	if (fseek(arch, offset, SEEK_SET) < 0) {
		return 1;
	}
	// Read header
	if (readULong(&header.size, arch) || readULong(&header.ctime, arch)
			|| readULong(&header.mtime, arch) || readULong(&header.hash, arch)) {
		return 1;
	}

	buffer = malloc(header.size);
	if (!buffer) {
		return 1;
	}

	if (fread(buffer, header.size, 1, arch) != 1) {
		return 1;
	}

	if (header.hash != crc32(buffer, header.size)) {
		printf("Warning: \"%s\" - bad hash.\n", fileName);
	}

	if ((buffer = decompress(buffer, &header.size)) == NULL) {
		return 1;
	}

	if (fwrite(buffer, header.size, 1, file) != 1) {
		return 1;
	}

	if (fclose(arch) == EOF || fclose(file) == EOF) {
		return 1;
	}

	return 0;
}

const char *archDeleteFiles(const char *archFileName, size_t count,
		char **names) {
	FILE *arch = fopen(archFileName, "rb");
	FILE *file = NULL;
	char *tempFileName = NULL;
	size_t i = 0;

	if (!arch) {
		return NULL;
	}

	do { // Make temp file name
		int size = 0, tempI = i;

		if (file) {
			fclose(file);
		}

		if (i == 256) {
			return NULL; // For not cycling
		}

		if (tempI == 0) {
			size = 1;
		} else {
			while (tempI) {
				tempI /= 10;
				size++;
			}
		}
		if (tempFileName) {
			free(tempFileName);
		}
		tempFileName = malloc(strlen(archFileName) + strlen(".swp") + size + 1);
		sprintf(tempFileName, "%s.swp%d", archFileName, i);

		file = fopen(tempFileName, "rb");

		i++;
	} while (file);

	file = fopen(tempFileName, "wb");

	if (!file) {
		return NULL;
	}

	{
		unsigned char byte;
		unsigned char *buffer = NULL;
		char name[1024];
		FileHeader header;

		if (fread(&byte, 1, 1, arch) != 1) {
			return NULL;
		}
		while (byte) {
			char nameByte;
			size_t nameLength = 0;
			char seek = 0;

			if (fread(&nameByte, 1, 1, arch) != 1) {
				return NULL;
			}
			while (nameByte) {
				name[nameLength++] = nameByte;
				if (fread(&nameByte, 1, 1, arch) != 1) { // Seek name
					return NULL;
				}
			}
			name[nameLength] = 0;

			for (i = 1; i < count; i++) {
				if (strcmp(names[i], name) == 0) {
					seek = 1;
					break;
				}
			}

			if (readULong(&header.size, arch) || readULong(&header.ctime, arch)
					|| readULong(&header.mtime, arch) || readULong(
					&header.hash, arch)) {
				return NULL;
			}

			if (seek) {
				if (fseek(arch, header.size, SEEK_CUR) < 0) {
					return NULL;
				}
			} else {
				if (fwrite(&byte, 1, 1, file) != 1) {
					return NULL;
				}
				if (fwrite(name, nameLength + 1, 1, file) != 1) {
					return NULL;
				}
				if (writeULong(header.size, file) || writeULong(header.ctime,
						file) || writeULong(header.mtime, file) || writeULong(
						header.hash, file)) {
					return NULL;
				}
				buffer = malloc(header.size);
				if (!buffer) {
					return NULL;
				}

				if (fread(buffer, header.size, 1, arch) != 1) {
					return NULL;
				}

				if (fwrite(buffer, header.size, 1, file) != 1) {
					return NULL;
				}

				free(buffer);
			}

			if (fread(&byte, 1, 1, arch) != 1) {
				return NULL;
			}
		}
		if (fwrite(&byte, 1, 1, file) != 1) {
			return NULL;
		}
	}

	if (fclose(arch) == EOF || fclose(file) == EOF) {
		return NULL;
	}

	return tempFileName;
}

void archPrintToScreen(const char *archFileName) {
	FILE *arch = fopen(archFileName, "rb");
	unsigned char byte;
	unsigned long totalCount = 0;
	unsigned long totalSize = 0;

	if (!arch) {
		return;
	}

	printf("Archive \"%s\" contain:\n\n", archFileName);

	if (fread(&byte, 1, 1, arch) != 1) {
		return;
	}
	while (byte) {
		char nameByte;
		unsigned long size;
		char name[1024];
		size_t nameLength = 0;

		if (fread(&nameByte, 1, 1, arch) != 1) {
			return;
		}
		while (nameByte) {
			name[nameLength++] = nameByte;
			if (fread(&nameByte, 1, 1, arch) != 1) { // Seek name
				return;
			}
		}
		name[nameLength] = 0;

		readULong(&size, arch);

		printf(" \"%s\" : %u\n", name, size);

		if (fseek(arch, size + sizeof(unsigned long) * 3, SEEK_CUR) < 0) {
			return;
		}

		totalCount++;
		totalSize += size;

		if (fread(&byte, 1, 1, arch) != 1) {
			return;
		}
	}

	printf("\nCompressing data: %u\n", totalSize);
	printf("Count of files: %u\n", totalCount);

	if (fclose(arch) == EOF) {
		return;
	}
}

unsigned long *archCreateOffsetTableForAll(const char *archFileName,
		size_t *count, char ***names) {
	int i;
	unsigned long *offsets = 0;
	int offset;
	char **newNames = 0;

	FILE *arch = fopen(archFileName, "rb");
	unsigned char byte;

	if (!arch) {
		return NULL;
	}

	if (fread(&byte, 1, 1, arch) != 1) {
		return NULL;
	}
	offset = 1;
	i = 0;
	while (byte) {
		char nameByte;
		unsigned long size;
		char name[1024];
		size_t nameLength = 0;

		offsets = realloc(offsets, (i + 1) * sizeof(unsigned long));
		newNames = realloc(newNames, (i + 1) * sizeof(char *));

		offset++;
		if (fread(&nameByte, 1, 1, arch) != 1) {
			return NULL;
		}
		while (nameByte) {
			name[nameLength++] = nameByte;
			offset++;
			if (fread(&nameByte, 1, 1, arch) != 1) { // Seek name
				return NULL;
			}
		}
		name[nameLength] = 0;
		offsets[i] = offset;
		newNames[i] = strdup(name);

		readULong(&size, arch);
		offset += size + sizeof(unsigned long) * 4;
		if (fseek(arch, size + sizeof(unsigned long) * 3, SEEK_CUR) < 0) {
			return NULL;
		}

		offset++;
		if (fread(&byte, 1, 1, arch) != 1) {
			return NULL;
		}
		i++;
	}

	if (fclose(arch) == EOF) {
		return NULL;
	}

	*names = newNames;
	*count = i;

	return offsets;
}


const char *archMakeTempAndFix(const char *archFileName) {
	FILE *file = fopen(archFileName, "rb");
	FILE *temp = NULL;
	char *tempFileName = NULL;
	int i = 0;

	if (!file) {
		return NULL;
	}

	do { // Make temp file name
		int size = 0, tempI = i;

		if (temp) {
			fclose(temp);
		}

		if (i == 256) {
			return NULL; // For not cycling
		}

		if (tempI == 0) {
			size = 1;
		} else {
			while (tempI) {
				tempI /= 10;
				size++;
			}
		}
		if (tempFileName) {
			free(tempFileName);
		}
		tempFileName = malloc(strlen(archFileName) + strlen(".swp") + size + 1);
		sprintf(tempFileName, "%s.swp%d", archFileName, i);

		temp = fopen(tempFileName, "rb");

		i++;
	} while (temp);

	temp = fopen(tempFileName, "wb");

	{ // Copy old archive to temp and stop when find a mistake
		unsigned char byte;
		unsigned char *buffer = NULL;
		char name[1024];
		FileHeader header;

		if (fread(&byte, 1, 1, file) != 1) {
			return NULL;
		}
		while (byte) {
			char nameByte;
			size_t nameLength = 0;

			if (fread(&nameByte, 1, 1, file) != 1) {
				return NULL;
			}
			while (nameByte) {
				name[nameLength++] = nameByte;
				if (fread(&nameByte, 1, 1, file) != 1) { // Seek name
					return NULL;
				}
			}
			name[nameLength] = 0;

			if (readULong(&header.size, file) || readULong(&header.ctime, file)
					|| readULong(&header.mtime, file) || readULong(
					&header.hash, file)) {
				return NULL;
			}

			buffer = malloc(header.size);
			if (!buffer) {
				return NULL;
			}

			if (fread(buffer, header.size, 1, file) != 1) {
				return NULL;
			}

			if (crc32(buffer, header.size) != header.hash) { // if bad hash then stop
				printf("Fixed: find a error on file \"%s\", delete the rest of archive.\n", name);
				break;
			}

			if (fwrite(&byte, 1, 1, temp) != 1) {
				return NULL;
			}
			if (fwrite(name, nameLength + 1, 1, temp) != 1) {
				return NULL;
			}
			if (writeULong(header.size, temp) || writeULong(header.ctime, temp)
					|| writeULong(header.mtime, temp) || writeULong(
					header.hash, temp)) {
				return NULL;
			}

			if (fwrite(buffer, header.size, 1, temp) != 1) {
				return NULL;
			}

			free(buffer);

			if (fread(&byte, 1, 1, file) != 1) {
				return NULL;
			}
		}
		byte = 0;
		if (fwrite(&byte, 1, 1, temp) != 1) {
			return NULL;
		}
	}
	fclose(file);
	fclose(temp);

	return tempFileName;
}

